(window.webpackJsonp=window.webpackJsonp||[]).push([[217],{1266:function(a,t){a.exports={attributes:{},body:'::::{tab-set}\n\n:::{tab-item} Feedback datasets\n\n```python\n# install datasets library with pip install datasets\nimport argilla as rg\nfrom datasets import load_dataset\n\n# load an Argilla Feedback Dataset from the Hugging Face Hub\ndataset = rg.FeedbackDataset.from_huggingface("argilla/oasst_response_quality", split="train")\n\n# push the dataset to Argilla\ndataset.push_to_argilla("oasst_response_quality")\n```\n:::\n\n:::{tab-item} Other datasets\n\n```python\n# install datasets library with pip install datasets\nimport argilla as rg\nfrom datasets import load_dataset\n\n# load dataset from the hub\ndataset = load_dataset("argilla/gutenberg_spacy-ner", split="train")\n\n# read in dataset, assuming its a dataset for token classification\ndataset_rg = rg.read_datasets(dataset, task="TokenClassification")\n\n# log the dataset\nrg.log(dataset_rg, "gutenberg_spacy-ner")\n```\n:::\n::::'}}}]);