(window.webpackJsonp=window.webpackJsonp||[]).push([[166],{1356:function(t,n,e){"use strict";e.r(n);e(8),e(176),e(35),e(177),e(178),e(179),e(180),e(181),e(182),e(183),e(184),e(185),e(186),e(187),e(188),e(189),e(190),e(191),e(192),e(30),e(29),e(23);var o={name:"TextClassificationBulkAnnotation",props:{datasetId:{type:Array,required:!0},records:{type:Array,required:!0},recordsIds:{type:Set,required:!0},labels:{type:Array,required:!0}},data:function data(){return{labelsByRecordId:[]}},computed:{formattedLabelsForBulkAnnotationForm:function formattedLabelsForBulkAnnotationForm(){return this.initLabelForForm()}},methods:{initLabelForForm:function initLabelForForm(){var t=this.labelsForFormFactory();return t=this.updateLabelByCurrentAnnotation(t)},labelsForFormFactory:function labelsForFormFactory(){return this.labels.map((function(t){return{record_ids:new Set,id:t,label:t,selected:!1,unmodified:!1,removed:!1}}))},updateLabelByCurrentAnnotation:function updateLabelByCurrentAnnotation(t){var n=this;return this.records.map((function(e){var o;null===(o=e.currentAnnotation)||void 0===o||o.labels.map((function(o){t.forEach((function(t){t.label===o.class&&(t.record_ids.add(e.id),t.selected=n.updateSelectedAttribute(t.record_ids.size))}))}))})),t},updateSelectedAttribute:function updateSelectedAttribute(t){return t===this.records.length},updateAnnotations:function updateAnnotations(t){this.$emit("on-update-annotations",t)}}},r=e(37),a=Object(r.a)(o,(function render(){var t=this,n=t._self._c;return n("div",[n("BulkAnnotationComponent",{staticClass:"bulk-annotation-component",attrs:{inputs:t.formattedLabelsForBulkAnnotationForm,recordsIds:t.recordsIds},on:{"on-update-annotations":t.updateAnnotations}})],1)}),[],!1,null,"993b2afc",null);n.default=a.exports;installComponents(a,{BulkAnnotationComponent:e(1357).default})}}]);