(window.webpackJsonp=window.webpackJsonp||[]).push([[206],{1008:function(a,e,t){"use strict";t.r(e),t.d(e,"useDatasetViewModel",(function(){return s}));t(19);var n=t(175),r=t(47),o=t(283),s=function useDatasetViewModel(){var a=Object(n.d)(!1),e=Object(n.h)().value.params.id;return{datasetId:e,isLoadingDataset:a,handleError:function handleError(a){var t="";switch(a){case o.a.ERROR_FETCHING_DATASET_INFO:t="Can't get dataset info for dataset_id: ".concat(e);break;case o.a.ERROR_FETCHING_WORKSPACE_INFO:t="Can't get workspace info for dataset_id: ".concat(e);break;default:t="There was an error on fetching dataset info and workspace info. Please try again"}var n={message:t,numberOfChars:t.length,type:"error"};r.a.dispatch("notify",n)},createRootBreadCrumbs:function createRootBreadCrumbs(a){return[{link:{name:"datasets"},name:"Home"},{link:{path:"/datasets?workspaces=".concat(a.workspace)},name:a.workspace},{link:{path:"annotation-mode"},name:a.name}]}}}}}]);