(window.webpackJsonp=window.webpackJsonp||[]).push([[229],{1278:function(i,t){i.exports={attributes:{title:"AutoTrain (no-code)",description:"Use Argilla, Datasets and Hugging Face AutoTrain with just a few clicks, empowering you to train NLP models easily without writing a single line of code.",links:[{linkText:"Tutorial",linkLink:"https://www.argilla.io/blog/argilla-meets-autotrain/"},{linkText:"AutoTrain UI",linkLink:"https://argilla-argilla-streamlit-customs.hf.space/autotrain-project"}]},body:"This is a no-code solution so take a look at the Tutorial and AutoTrain UI links underneath."}}}]);