(window.webpackJsonp=window.webpackJsonp||[]).push([[33],{968:function(e,t,n){"use strict";n.r(t);var r={props:{date:{type:String},format:{type:String}},computed:{formattedDate:function formattedDate(){var e=new Date(this.date);return"date-relative-now"===this.format?this.timeAgo(e):e.toLocaleString("sv",{year:"numeric",month:"numeric",day:"numeric",hour:"numeric",minute:"numeric",hour12:!1})}},methods:{timeAgo:function timeAgo(e){var t=new Intl.RelativeTimeFormat("en",{numeric:"auto"}),n={years:31536e3,months:2592e3,weeks:604800,days:86400,hours:3600,minutes:60,seconds:1},r=new Date,a=new Date(e.getTime()-r.getTime());a.setMinutes(a.getMinutes()-r.getTimezoneOffset());var i=a/1e3;for(var o in n)if(n[o]<=Math.abs(i)){var s=i/n[o];return t.format(Math.round(s),o)}}}},a=n(37),i=Object(a.a)(r,(function render(){var e=this;return(0,e._self._c)("span",[e._v(" "+e._s(e.formattedDate)+" ")])}),[],!1,null,null,null);t.default=i.exports}}]);