(window.webpackJsonp=window.webpackJsonp||[]).push([[210],{1284:function(e,t,n){"use strict";n.r(t),n.d(t,"useDatasetsViewModel",(function(){return o}));n(16),n(8);var a=n(48),c=n(175),r=n(297),u=n(295),__awaiter=function(e,t,n,a){return new(n||(n=Promise))((function(c,r){function fulfilled(e){try{step(a.next(e))}catch(e){r(e)}}function rejected(e){try{step(a.throw(e))}catch(e){r(e)}}function step(e){e.done?c(e.value):function adopt(e){return e instanceof n?e:new n((function(t){t(e)}))}(e.value).then(fulfilled,rejected)}step((a=a.apply(e,t||[])).next())}))},o=function useDatasetsViewModel(){var e=Object(c.d)(!1),t=Object(u.a)().state,n=Object(a.useResolve)(r.a);Object(c.g)((function(){o()}));var o=function loadDatasets(){return __awaiter(void 0,void 0,void 0,regeneratorRuntime.mark((function _callee(){return regeneratorRuntime.wrap((function _callee$(t){for(;;)switch(t.prev=t.next){case 0:return e.value=!0,t.next=3,n.execute();case 3:e.value=!1;case 4:case"end":return t.stop()}}),_callee)})))};return{datasets:t,isLoadingDatasets:e}}}}]);