(window.webpackJsonp=window.webpackJsonp||[]).push([[219],{1268:function(e,i){e.exports={attributes:{title:"Supervised Fine Tuning",description:"The goal of Supervised Fine Tuning (SFT) is to optimize this pre-trained model to generate the responses that users are looking for. As with RLHF we don't provide direct integrations yet but we do offer a general guide to get you started.",links:[{linkText:"Practical guide to SFT",linkLink:"https://docs.argilla.io/en/latest/guides/llms/practical_guides/fine_tune.html#supervised-finetuning"}]},body:""}}}]);