(window.webpackJsonp=window.webpackJsonp||[]).push([[159],{1176:function(e,n,t){"use strict";t.r(n);var o={props:{options:{type:Array,required:!0},selectedOption:{type:String}},model:{prop:"selectedOption",event:"change"},methods:{onChangeOption:function onChangeOption(e){this.$emit("change",e)}}},i=t(37),s=Object(i.a)(o,(function render(){var e=this,n=e._self._c;return e.options.length?n("div",[n("RadioButtonsSelectBase",{attrs:{options:e.options,"selected-option":e.selectedOption},on:{change:e.onChangeOption}})],1):e._e()}),[],!1,null,null,null);n.default=s.exports;installComponents(s,{RadioButtonsSelectBase:t(1161).default})}}]);