#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from argilla.client.feedback import (
    ArgillaTrainer,
    LabelQuestionStrategy,
    LabelQuestionUnification,
    MultiLabelQuestionStrategy,
    MultiLabelQuestionUnification,
    RankingQuestionStrategy,
    RankingQuestionUnification,
    RatingQuestionStrategy,
    RatingQuestionUnification,
    TrainingTask,
    TrainingTaskMapping,  # <- Deprecated
)
from argilla.client.feedback.dataset import FeedbackDataset
from argilla.client.feedback.schemas import (
    FeedbackRecord,
    LabelQuestion,
    MultiLabelQuestion,
    RankingQuestion,
    RatingQuestion,
    ResponseSchema,
    SuggestionSchema,
    TextField,
    TextQuestion,
    ValueSchema,
)

__all__ = [
    "ArgillaTrainer",
    "LabelQuestionStrategy",
    "MultiLabelQuestionStrategy",
    "RatingQuestionStrategy",
    "RankingQuestionStrategy",
    "LabelQuestionUnification",
    "MultiLabelQuestionUnification",
    "RatingQuestionUnification",
    "RankingQuestionUnification",
    "TrainingTask",
    "TrainingTaskMapping",
    "FeedbackDataset",
    "FeedbackRecord",
    "LabelQuestion",
    "MultiLabelQuestion",
    "RatingQuestion",
    "RankingQuestion",
    "ResponseSchema",
    "SuggestionSchema",
    "TextField",
    "TextQuestion",
    "ValueSchema",
]
